#pragma once

typedef int RwCamera;
typedef int RwTexture;
typedef int RwRaster;
typedef char Bool;
typedef int RpLight;
typedef int RwInt32;
typedef struct RwObjectHasFrame RwObjectHasFrame;

struct RwRGBA
{
	char r;
	char g;
	char b;
	char a;
};

#pragma pack(push, 1)
struct RwObject
{
  char type;
  char subType;
  char flags;
  char privateFlags;
  void *parent;
};
#pragma pack(pop)

struct RwSphere
{
  float x;
  float y;
  float z;
  float radius;
};

struct RwV3d
{
  float x;
  float y;
  float z;
};

struct RwMatrix
{
  RwV3d right;
  int flags;
  RwV3d top;
  int pad1;
  RwV3d at;
  int pad2;
  RwV3d pos;
  int pad3;
};

struct RwLLLink
{
  int next;
  int prev;
};

struct RwLinkList
{
  RwLLLink link;
};

struct RwFrame
{
  RwObject object;
  RwLLLink inDirtyListLink;
  RwMatrix modelling;
  RwMatrix ltm;
  RwLinkList objectList;
  int child;
  int next;
  int root;
};

struct RpInterpolator
{
  int flags;
  short startMorphTarget;
  short endMorphTarget;
  int time;
  int recipTime;
  int position;
};

struct RwObjectHasFrame
{
  RwObject object;
  RwLLLink lFrame;
  int sync;
};

struct RpAtomic
{
  RwObjectHasFrame object;
  void *repEntry;
  void *geometry;
  RwSphere boundingSphere;
  RwSphere worldBoundingSphere;
  void *clump;
  RwLLLink inClumpLink;
  int renderCallback;
  RpInterpolator interpolator;
  unsigned __int16 renderFrame;
  unsigned __int16 pad;
  RwLinkList llWorldSectorsInAtomic;
  void *pipeline;
};

class CVector
{
public:
	float x;
	float y;
	float z;
	
	CVector () { }

	CVector(float X, float Y, float Z) : x(X), y(Y), z(Z) 
	{}
	
	float Normalise()
	{
		return ((float(__thiscall *)(CVector *))0x4DFEA0)(this);
	}
};

class CVector2D
{
public:
	float x;
	float y;
	
	CVector2D(float X, float Y) : x(X), y(Y)
	{}
};

class CMatrix
{
public:
	union
	{
		struct
		{
			// RwV3d-like:
			CVector		right;
			unsigned int flags;
			CVector		up;
			unsigned int pad1;
			CVector		at;
			unsigned int pad2;
			CVector		pos;
			unsigned int pad3;
		};
		struct
		{
			RwMatrix  matrix;     //!< +0x00
		};
	};

	void *m_pAttachMatrix;
	unsigned int m_bAttachMatrixTemporary; // do we need to delete attaching matrix at detaching
	
	CMatrix(CMatrix const& matrix)
	{
		((void (__thiscall *)(CMatrix *, CMatrix const&))0x4DFBA0)(this, matrix);
	}
	
};

class CPlaceable
{
public:
	void *vtable;
	CMatrix Coords;
};

class CEntity : public CPlaceable
{
public:
	union {
		void *m_pRwObject;
		void *m_pAtomic;
		void	*m_pClump;
	};

	unsigned char m_nType : 3;
	unsigned char m_nState : 5;

	struct {
		unsigned char bUseCollision : 1;
		unsigned char bEntUFlag02 : 1;
		unsigned char bIsStatic : 1;
		unsigned char bEntUFlag04 : 1;
		unsigned char bEntUFlag05 : 1;
		unsigned char bEntUFlag06 : 1;
		unsigned char bEntUFlag07 : 1;
		unsigned char bRecordCollisions : 1;

		unsigned char bEntUFlag09 : 1;
		unsigned char bExplosionProof : 1;
		unsigned char bIsVisible : 1;
		unsigned char bHasCollided : 1;
		unsigned char bRenderScorched : 1;
		unsigned char bEntUFlag14 : 1;
		unsigned char bUseLevelSectors : 1;
		unsigned char bIsBigBuilding : 1;

		unsigned char bEntUFlag17 : 1;
		unsigned char bBulletProof : 1;
		unsigned char bFireProof : 1;
		unsigned char bCollisionProof : 1;
		unsigned char bEntUFlag21 : 1;
		unsigned char bEntUFlag22 : 1;
		unsigned char bEntUFlag23 : 1;
		unsigned char bRemoveFromWorld : 1;

		unsigned char bEntUFlag25 : 1;
		unsigned char bImBeingRendered : 1;
		unsigned char bEntUFlag27 : 1;
		unsigned char bEntUFlag28 : 1;
		unsigned char bEntUFlag29 : 1;
		unsigned char bEntUFlag30 : 1;
		unsigned char bEntUFlag31 : 1;
		unsigned char bEntUFlag32 : 1;

		unsigned char bEntUFlag33 : 1;
		unsigned char bDontCastShadowsOn : 1;
		unsigned char bEntUFlag35 : 1;
		unsigned char bIsStaticWaitingForCollision : 1;
		unsigned char bEntUFlag37 : 1;
		unsigned char bEntUFlag38 : 1;
		unsigned char bEntUFlag39 : 1;
		unsigned char bEntUFlag40 : 1;
	} m_nFlags;

	char _pad[6];

	short m_nModelIndex;
	char m_nLevel;
	unsigned char m_nInterior;
	void *m_pFirstRef;
};

class CEntryInfoNode
{
public:
	void *m_pPtrList;
	void *m_pPtrNode;
	void *m_pSector;
	CEntryInfoNode *m_pNext;
	CEntryInfoNode *m_pPrev;
};

class CEntryInfoList
{
public:
		CEntryInfoNode *m_pLastEntry;
};

class CPhysical : public CEntity
{
public:
	unsigned int m_nAudioEntityId;
	float field_68;
	int field_6C;
	CVector m_vecMoveSpeed;
	CVector m_vecTurnSpeed;
	CVector m_vecAcceleration;
	CVector m_vecAngularAcceleration;
	CVector m_vecForce;
	CVector m_vecTorque;
	float m_fMass;
	float m_fTurnMass;
	float m_fVelocityFrequency;
	float m_fAirResistance;
	float m_fElasticity;
	float m_fBuoyancyConstant;
	CVector m_vecCentreOfMass;
	CEntryInfoList m_collisionList;
	void *m_pMovingListNode;
	char field_E4[2];
	unsigned char m_nNumCollisionRecords;
	char field_E7;
	CEntity *m_apCollisionRecords[6];
	float m_fTotSpeed;
	float m_fCollisionPower;
	CPhysical *m_pPhysColliding;
	CVector m_vecCollisionPower;
	unsigned short m_wComponentCol;
	unsigned char m_nMoveFlags;
	unsigned char m_nCollFlags;
	unsigned char m_nLastCollType;
	unsigned char m_nZoneLevel;
	char field_11E[2];
};

class CCollPoly
{
public:
	CVector vecColPolyPoint1;
	CVector vecColPolyPoint2;
	CVector vecColPolyPoint3;
	char bIsValidCollision;
	char __fx0025[3];
};

class CPedIK
{
public:
	int pPed;
	char __f0004[32];
	int dwFlags;
};

class CPathNode
{
public:
	__int16 wPathData1;
	__int16 wPathData2;
	__int16 wPosX;
	__int16 wPosY;
	__int16 wPosZ;
	char __f000A[10];
};

class CWeapon
{
public:
	int nWeaponId;
	int nWeaponState;
	int nAmmoLoaded;
	int nAmmoTotal;
	int dwNextShotTime;
	char bAddRotOffset;
	char __f0015[3];
};

class CPed : public CPhysical
{
public:
	CCollPoly polyColliding;
	int fCollisionSpeed;	
	struct  {
		unsigned char bIsStanding : 1;
		unsigned char bWasStanding : 1;
		unsigned char bPedFlags03 : 1;
		unsigned char bIsPointingGunAt : 1;
		unsigned char bIsLooking : 1;
		unsigned char bPedFlags06 : 1;
		unsigned char bIsRestoringLook : 1;
		unsigned char bIsAimingGun : 1;

		unsigned char bIsRestoringGun : 1;
		unsigned char bCanPointGunAtTarget : 1;
		unsigned char bIsTalking : 1;
		unsigned char bIsInTheAir : 1;
		unsigned char bIsLanding : 1;
		unsigned char bIsRunning : 1;
		unsigned char bPedFlags15 : 1;
		unsigned char bPedFlags16 : 1;

		unsigned char bCanPedEnterSeekedCar : 1;
		unsigned char bRespondsToThreats : 1;
		unsigned char bRenderPedInCar : 1;
		unsigned char bPedFlags20 : 1;
		unsigned char bUpdateAnimHeading : 1;
		unsigned char bRemoveHead : 1;
		unsigned char bFiringWeapon : 1;
		unsigned char bPedFlags24 : 1;

		unsigned char bPedFlags60a : 1;
		unsigned char bPedFlags60b : 1;
		unsigned char bPedIsBleeding : 1;
		unsigned char bStopAndShoot : 1;
		unsigned char bIsPedDieAnimPlaying : 1;
		unsigned char bPedFlags30 : 1;
		unsigned char bPedFlags31 : 1;
		unsigned char bPedFlags32 : 1;

		unsigned char bKindaStayInSamePlace : 1;
		unsigned char bBeingChasedByPolice : 1;
		unsigned char bNotAllowedToDuck : 1;
		unsigned char bCrouchWhenShooting : 1;
		unsigned char bIsDucking : 1;
		unsigned char bGetUpAnimStarted : 1;
		unsigned char bDoBloodyFootprints : 1;
		unsigned char bPedFlags40 : 1;

		unsigned char bPedFlags41 : 1;
		unsigned char bPedFlags42 : 1;
		unsigned char bPedFlags43 : 1;
		unsigned char bPedFlags44 : 1;
		unsigned char bPedFlags45 : 1;
		unsigned char bPedFlags46 : 1;
		unsigned char bPedFlags47 : 1;
		unsigned char bPedFlags48 : 1;

		unsigned char bPedFlags49 : 1;
		unsigned char bPedFlags50 : 1;
		unsigned char bPedFlags51 : 1;
		unsigned char bPedFlags52 : 1;
		unsigned char bPedFlags53 : 1;
		unsigned char bPedFlags54 : 1;
		unsigned char bIsBeingJacked : 1;
		unsigned char bFadeOut : 1;

		unsigned char bPedFlags57 : 1;
		unsigned char bPedFlags58 : 1;
		unsigned char bPedFlags59 : 1;
		unsigned char bClearObjective : 1;
		unsigned char bPedFlags61 : 1;
		unsigned char bPedFlags62 : 1;
		unsigned char bPedFlags63 : 1;
		unsigned char bPedFlags64 : 1;

		unsigned char bShakeFist : 1;
		unsigned char bNoCriticalHits : 1;
		unsigned char bPedFlags67 : 1;
		unsigned char bHasAlreadyBeenRecorded : 1;
		unsigned char bFallenDown : 1;
		unsigned char bUpdateMatricesRequired : 1;
		unsigned char bPedFlags71 : 1;
		unsigned char bPedFlags72 : 1;

		unsigned char bPedFlags73 : 1;
		unsigned char bPedFlags74 : 1;
		unsigned char bPedFlags75 : 1;
		unsigned char bPedFlags76 : 1;
		unsigned char bPedFlags77 : 1;
		unsigned char bMiamiViceCop : 1;
		unsigned char bMoneyHasBeenGivenByScript : 1;
		unsigned char bHasBeenPhotographed : 1;

		unsigned char bIsDrowning : 1;
		unsigned char bDrownsInWater : 1;
		unsigned char bStayBehindIfLeaderTooFar : 1;
		unsigned char bUsingExitCarTimer : 1;
		unsigned char bScriptPedIsPlayerAlly : 1;
		unsigned char bKnockedUpIntoAir : 1;
		unsigned char bDeadPedInFrontOfCar : 1;
		unsigned char bStayInCarOnJack : 1;

		unsigned char bPedFlags89 : 1;
		unsigned char bDoomAim : 1;
		unsigned char bCanBeShotInVehicle : 1;
		unsigned char bPedFlags92 : 1;
		unsigned char bPedFlags93 : 1;
		unsigned char bPedFlags94 : 1;
		unsigned char bPedFlags95 : 1;
		unsigned char bPedFlags96 : 1;

		unsigned char bPedFlags97 : 1;
		unsigned char bCrouchWhenScared : 1;
		unsigned char bPedFlags99 : 1;
		unsigned char bPedFlags100 : 1;
		unsigned char bPedFlags101 : 1;
		unsigned char bPedFlags102 : 1;
		unsigned char bPedWasSetOutOfCollision : 1;
		unsigned char bGangMemberReturnsFire : 1;
	
	}PedFlags;
	
	char __f0155[3];
	char bfGangFlags;
	char __f0159[3];
	char charPedStatus;
	char __fx015D[3];
	int dwObjective;
	int dwPrevObjective;
	CEntity *pObjectiveEntity;
	void* pObjectiveVehicle;
	CVector vecObjective;
	float fObjectiveAngle;
	CPed *pGangLeader;
	int dwPedFormation;
	int dwFearFlags;
	CEntity * pThreatEntity;
	float fEventOrThreatX;
	float fEventOrThreatY;
	int dwEventType;
	CEntity * pEventEntity;
	float fAngleToEvent;
	int pFrames[18];
	void* pCurWeaponAtomic;
	int dwAnimGroupId;
	int pVehicleAnim;
	CVector2D vecAnimMoveDelta;
	CVector vecOffsetSeek;
	CPedIK stPedIK;
	float fActionX;
	float fActionY;
	int dwActionTimer;
	int dwAction;
	int dwLastAction;
	int dwMoveState;
	int dwStoredActionState;
	int dwPrevActionState;
	int dwWaitState;
	int dwWaitTimer;
	int pPathNodesStates[8];
	__int16 wPathNodes;
	__int16 wCurPathNode;
	CEntity * pPathRelEntity;
	CEntity * pNextNodeEntity;
	int dwPathNodeTimer;
	CPathNode stPathNodeStates[8];
	int pCurNodeState;
	char charPathState;
	char __f0331[3];
	CPathNode* pNextPathNode;
	CPathNode* pLastPathNode;
	CVector vecPathNextNode;
	float fPathNextNodeDir;
	int dwPathNodeType;
	float fHealth;
	float fArmour;
	int dwShadowUpdateTimer;
	__int16 wRouteLastPoint;
	__int16 wRoutePoints;
	__int16 wRoutePos;
	__int16 wRouteType;
	__int16 wRouteCurDir;
	char __f0366[2];
	float fMovedX;
	float fMovedY;
	float fRotationCur;
	float fRotationDest;
	float fHeadingRate;
	__int16 wEnterType;
	__int16 wWalkAroundType;
	void* pCurPhysSurface;
	CVector vecOffsetFromPhysSurface;
	void* pCurSurface;
	CVector vecSeekVehicle;
	CEntity *pSeekTarget;
	CEntity *pVehicle;
	char charIsInVehicle;
	char __f03A9[3];
	float fSeatPrecisionX;
	float fSeatPrecisionY;
	void* pFromVehicle;
	void*	pSeat;
	int dwSeatType;
	char charHasPhone;
	char __f03C1;
	__int16 wPhoneId;
	int dwLookingForPhone;
	int dwPhoneTalkTimer;
	int pLastAccident;
	int dwPedType;
	int pPedStats;
	float fFleeFromPosX;
	float fFleeFromPosY;
	void* pFleeFrom;
	int dwFleeTimer;
	CEntity* pThreatEx;
	CEntity* pLastThreatAt;
	int dwLastThreatTimer;
	int pVehicleColliding;
	char charStateUnused;
	char __f03F9[3];
	struct CCutsceneShadow *dwTimerUnused;
	void * pTargetUnused;
	CWeapon stWeps[10];
	int dwAtchStoredWep;
	int dwStoredGiveWep;
	int dwStoredGiveAmmo;
	char charWepSlot;
	char charWepSkills;
	char charWepAccuracy;
	char charBodyPart;
	int pPointGunAt;
	CVector vecHitLastPos;
	int dwHitCounter;
	int dwLastHitState;
	char charFightFlags1;
	char charFightFlags2;
	char charFightFlags3;
	char charBleedCounter;
	void* pPedFire;
	CPed* pPedFight;
	float fLookDirection;
	int dwWepModelID;
	int dwLeaveCarTimer;
	int dwGetUpTimer;
	int dwLookTimer;
	int dwStandardTimer;
	int dwAttackTimer;
	int dwLastHitTime;
	int dwHitRecoverTimer;
	int dwObjectiveTimer;
	int dwDuckTimer;
	int dwDuckAndCoverTimer;
	int dwBloodyTimer;
	int dwShotTime;
	int dwShotTimeAdd;
	char charPanicCounter;
	char charDeadBleeding;
	char charBodyPartBleeding;
	char __f0567;
	CPed *pNearPeds[10];
	__int16 nNearPeds;
	__int16 wPedMoney;
	char charLastDamWep;
	char __f0595[3];
	int pLastDamEntity;
	int pAttachedTo;
	CVector vAtchOff;
	__int16 wAtchType;
	char _f05AE[2];
	float fAtchRot;
	int dwAtchWepAmmo;
	int dwThreatFlags;
	int dwThreatCheck;
	int dwLastThreatCheck;
	int dwSayType;
	int dwSayTimer;
	int dwTalkTimerLast;
	int dwTalkTimer;
	__int16 wTalkTypeLast;
	__int16 wTalkType;
	char charCanPedTalk;
	char __f05D9[3];
	int dwPedLastComment;
	CVector vecSeekPosEx;
	float fSeekExAngle;
};

struct CQueuedCamMode
{
  int Mode;
  float Duration;
  int MinZoom;
  int MaxZoom;
};

struct CCam
{
  char bBelowMinDist;
  char bBehindPlayerDesired;
  char bCamLookingAtVector;
  char bCollisionChecksOn;
  char bFixingBeta;
  char bTheHeightFixerVehicleIsATrain;
  char bLookBehindCamWasInFront;
  char bLookingBehind;
  char bLookingLeft;
  char bLookingRight;
  char bRotating;
  char _pad1[1];
  __int16 nMode;
  char _pad2[2];
  int dwFinishTime;
  int nDoCollisionChecksOnFrameNum;
  int nDoCollisionCheckEveryNumOfFrames;
  int nFrameNumWereAt;
  int nRunningVectorArrayPos;
  int nRunningVectorCounter;
  int nDirectionWasLooking;
  float fMaxRollAngle;
  float fRoll;
  float fRollSpeed;
  float fSyphonModeTargetZOffSet;
  float fAmountFractionObscured;
  float fAlphaSpeedOverOneFrame;
  float fBetaSpeedOverOneFrame;
  float fBufferedTargetBeta;
  float fBufferedTargetOrientation;
  float fBufferedTargetOrientationSpeed;
  float fCamBufferedHeight;
  float fCamBufferedHeightSpeed;
  float fCloseInPedHeightOffset;
  float fCloseInPedHeightOffsetSpeed;
  float fCloseInCarHeightOffset;
  float fCloseInCarHeightOffsetSpeed;
  float fDimensionOfHighestNearCar;
  float fDistanceBeforeChanges;
  float fFovSpeedOverOneFrame;
  float fMinDistAwayFromCamWhenInterPolating;
  float fPedBetweenCameraHeightOffset;
  float fPlayerInFrontSyphonAngleOffset;
  float fRadiusForDead;
  float fRealGroundDist;
  float fTargetBeta;
  float fTimeElapsedFloat;
  float fTilt;
  float fTiltSpeed;
  float fTransitionBeta;
  float fTrueBeta;
  float fTrueAlpha;
  float fInitialPlayerOrientation;
  float fAlpha;
  float fAlphaSpeed;
  float fFOV;
  float fFOVSpeed;
  float fBeta;
  float fBetaSpeed;
  float fDistance;
  float fDistanceSpeed;
  float fCaMinDistance;
  float fCaMaxDistance;
  float fSpeedVar;
  float fTargetZoomGroundOne;
  float fTargetZoomGroundTwo;
  float fTargetZoomGroundThree;
  float fTargetZoomOneZExtra;
  float fTargetZoomTwoZExtra;
  float fTargetZoomThreeZExtra;
  float fTargetZoomZCloseIn;
  float fMinRealGroundDist;
  float fTargetCloseInDist;
  CVector vecSourceSpeedOverOneFrame;
  CVector vecTargetSpeedOverOneFrame;
  CVector vecUpOverOneFrame;
  CVector vecTargetCoorsForFudgeInter;
  CVector vecCamFixedModeVector;
  CVector vecCamFixedModeSource;
  CVector vecCamFixedModeUpOffSet;
  CVector vecLastAboveWaterCamPosition;
  CVector vecBufferedPlayerBodyOffset;
  CVector vecLookFront;
  CVector vecSource;
  CVector vecSourceBeforeLookBehind;
  CVector vecLookUp;
  CVector vecArrPrevVectors[2];
  CEntity *pTargetEntity;
  float fCameraDistance;
  float fIdealAlpha;
  float fPlayerVelocity;
  void *pLastCarEntered;
  CPed *pLastPedLookedAt;
  char bFirstPersonRunAboutActive;
  char _pad0[3];
};

#pragma pack(push, 8)
struct CCamera
{
  CMatrix Coords;
  char _pad0[2];
  char bFollowPlayer;
  char _pad1[34];
  char bWideScreen;
  char _pad2[8];
  char charActiveCam;
  char _pad3[273];
  CCam stCams[3];
  CQueuedCamMode stPlayerMode;
  CQueuedCamMode stPlayerWeaponMode;
  CVector vecPrevCamPos;
  CVector vecRealPrevCamPos;
  CVector vecAimingTargetCoors;
  CVector vecFixedModeVector;
  CVector vecFixedModeSource;
  CVector vecFixedModeUpOffset;
  CVector vecCutSceneOffset;
  CVector vecStartingSourceForInterPol;
  CVector vecStartingTargetForInterPol;
  CVector vecStartingUpForInterPol;
  CVector vecSourceSpeedAtStartInter;
  CVector vecTargetSpeedAtStartInter;
  CVector vecUpSpeedAtStartInter;
  CVector vecSourceWhenInterPol;
  CVector vecTargetWhenInterPol;
  CVector vecUpWhenInterPol;
  CVector vecClearGeometry;
  CVector vecGameCamPos;
  CVector vecSourceDuringInter;
  CVector vecTargetDuringInter;
  CVector vecUpDuringInter;
  char __pad06[24];
  CMatrix matView;
  char _pad06[8];
  CMatrix matWorld;
  CVector m_avFrustumNormals[4];
  CVector field_8F0;
  CVector field_8FC;
  CVector field_908;
  char _unk[12];
  float m_fFadeDuration;
  float field_91C;
  char gap920[32];
  __int16 m_nFadingDirection;
  char _pad9[2];
  int m_nModeObbeCamIsInForCar;
  __int16 m_nModeToGoTo;
  __int16 m_nMusicFadingDirection;
  __int16 m_nTypeOfSwitch;
  char _pad4[2];
  int m_nFadeTimeStarted;
  int m_nFadeTimeStartedMusic;
};
#pragma pack(pop)